//-----------------------------------------------------------------------------
// Project            	RenderKit
// Class		G3DASELoader
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Mon May  8 10:17:53 CEST 2000
//
// Copyright (c) Philippe C.D. Robert
//
// The RenderKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DASELoader.m,v 1.1 2001/07/15 14:03:37 probert Exp $
//-----------------------------------------------------------------------------

#import "G3DASELoader.h"
#import "G3DASEObject.h"
#import "G3DVtxArray+ASEExtensions.h"

#import <RenderKit/RenderKit.h>

#define ASE_VERSION        @"*3DSMAX_ASCIIEXPORT"
#define ASE_COMMENT        @"*COMMENT"
#define ASE_MATERIALS      @"*MATERIAL_LIST"
#define ASE_MATERIAL       @"*MATERIAL"
#define ASE_MATERIAL_COUNT @"*MATERIAL_COUNT"
#define ASE_GEOMOBJC       @"*GEOMOBJECT"
#define ASE_MESH_NORMALS   @"*MESH_NORMALS"
#define ASE_MESH_VNORMAL   @"*MESH_VERTEXNORMAL"
#define ASE_MESH_FNORMAL   @"*MESH_FACENORMAL"

#define VALID_VERSION 200
#define RAISE(x) [NSException raise:@"G3DASELoaderException" format:(x)]
#define STEP_BACK(x) [_ASE_Scanner setScanLocation:[_ASE_Scanner scanLocation]-(x)]

NSString *G3DASELoaderException = @"G3DASELoaderException";

static NSScanner *_ASE_Scanner = nil;

@interface G3DASELoader (ASEPrivateMethods)

- (G3DGraphNode *)_parseASEFile;

- (void)_scanVersionInformation;
- (void)_scanCommentInformation;
- (void)_scanSceneInformation;

- (void)_scanMaterialList;
- (void)_scanMaterialInformation;

- (void)_scanGeometryObjectAtLocation:(unsigned)loc;
- (void)_scanNodeTransformation:(float *)arr;
- (G3DASEObject *)_scanMeshData;
- (void)_scanVertexList:(G3DASEObject **)data;
- (void)_scanFaceList:(G3DASEObject **)data;
- (void)_scanMeshNormals:(G3DASEObject **)data;

- (void)_scanLineNamed:(NSString *)nm intoString:(NSString **)dest;
- (void)_scanFloatsNamed:(NSString *)nm intoArray:(float *)arr count:(int)cnt;
- (void)_scanIntsNamed:(NSString *)nm intoArray:(int *)arr count:(int)cnt;

- (G3DGraphNode *)_createObjectTree;

@end

@implementation G3DASELoader (ASEPrivateMethods)

- (G3DGraphNode *)_parseASEFile
{
  /*
   * Algorithm: Parsing the version number - must be 200
   *            Parsing the SCENE information
   *            Parsing the material list
   *            Parsing the geometry objects
   *            Initialising the appropriate graphnodes and attributes
   *
   */

  [self _scanVersionInformation];
  [self _scanCommentInformation];
  [self _scanSceneInformation];
  [self _scanMaterialList];

  while ([_ASE_Scanner isAtEnd] == NO) {
    if ([_ASE_Scanner scanUpToString:ASE_GEOMOBJC intoString:NULL] &&
	[_ASE_Scanner scanString:ASE_GEOMOBJC intoString:NULL]) {
      [self _scanGeometryObjectAtLocation:[_ASE_Scanner scanLocation]];
    }
  }

  [aseFile autorelease];
  
  return [self _createObjectTree];
}

- (void)_scanVersionInformation
{
  int version;

  while ([_ASE_Scanner isAtEnd] == NO) {
    if ([_ASE_Scanner scanString:ASE_VERSION intoString:NULL] &&
	[_ASE_Scanner scanInt:&version]) {

      DLOG(@"<%@ %x>: ASE version %d",[self class],self,version);

      if (version != VALID_VERSION) {
	[NSException raise:G3DASELoaderException 
		     format:@"Version %d not valid!",version];
      }
      return;
    }
  }
}

- (void)_scanCommentInformation
{
  int loc = [_ASE_Scanner scanLocation];
  NSString *comment;

  STEP_BACK(1);

  while ([_ASE_Scanner isAtEnd] == NO) {
    if ([_ASE_Scanner scanUpToString:ASE_COMMENT intoString:NULL] &&
	[_ASE_Scanner scanString:ASE_COMMENT intoString:NULL]) {
      if ([_ASE_Scanner scanUpToString:@"\n" intoString:&comment]) {
	DLOG(@"<%@ %x>: ASE comment %@",[self class],self,comment);	
	STEP_BACK(1);
      }
    }
  }
  [_ASE_Scanner setScanLocation:loc];
}

- (void)_scanSceneInformation
{
  DLOG(@"<%@ %x>: ASE scene ignored...",[self class],self);
}

- (void)_scanMaterialList
{
  int materialCount;

  STEP_BACK(1);

  while ([_ASE_Scanner isAtEnd] == NO) {
    if ([_ASE_Scanner scanUpToString:ASE_MATERIAL_COUNT intoString:NULL] &&
	[_ASE_Scanner scanString:ASE_MATERIAL_COUNT intoString:NULL] &&
	[_ASE_Scanner scanInt:&materialCount]) {

      DLOG(@"<%@ %x>: ASE material count %d",[self class],self,materialCount);

      if (materialCount > 0) {
	materials = [NSMutableDictionary dictionaryWithCapacity:materialCount];

	[self _scanMaterialInformation];
      }
      else {
	DLOG(@"No material information found!");
      }
      return;
    }
  }
}

- (void)_scanMaterialInformation
{
  int loc = [_ASE_Scanner scanLocation];
  int number;

  STEP_BACK(1);

  while ([_ASE_Scanner isAtEnd] == NO) {
    if ([_ASE_Scanner scanUpToString:ASE_MATERIAL intoString:NULL] &&
	[_ASE_Scanner scanString:ASE_MATERIAL intoString:NULL]) {

      if ([_ASE_Scanner scanInt:&number]) {
	G3DMaterial *_mat = [G3DMaterial materialWithState:nil];
	NSString *tmp = [_mat name];
	float *arr;
	float _sh;
	
	[self _scanLineNamed:@"*MATERIAL_NAME" intoString:&tmp];
	
	arr = [[_mat ambientColour] values];
	[self _scanFloatsNamed:@"*MATERIAL_AMBIENT" intoArray:arr count:3];

	arr = [[_mat diffuseColour] values];
	[self _scanFloatsNamed:@"*MATERIAL_DIFFUSE" intoArray:arr count:3];

	arr = [[_mat specularColour] values];
	[self _scanFloatsNamed:@"*MATERIAL_SPECULAR" intoArray:arr count:3];

	[self _scanFloatsNamed:@"*MATERIAL_SHINE" intoArray:&_sh count:1];
	[_mat setShininess:(double)_sh];

	DLOG(@"<%@ %x>: ASE Material %d...",[self class],self,number);	
	
	[materials setObject:_mat forKey:[NSNumber numberWithInt:number]];

	STEP_BACK(1);
      }
    }
  }
  [_ASE_Scanner setScanLocation:loc];
}

- (void)_scanGeometryObjectAtLocation:(unsigned)loc
{
  G3DASEObject *obj;
  NSString *name;
  float m[16];

  if (!objects) {
    objects = [NSMutableDictionary dictionary];
  }

  [self _scanLineNamed:@"*NODE_NAME" intoString:&name];
  DLOG(@"<%@ %x>: Geometry object %@ at %d",[self class],self,name,loc);

  [self _scanNodeTransformation:m];

  if ((obj = [self _scanMeshData])) {
    [obj setName:name];
    [obj setTransformationWithElements:m];

    // Add the object to the dictionary    
    [objects setObject:obj forKey:name];
  }

  // this is needed to find the next geometry object...
  [_ASE_Scanner setScanLocation:loc+1];
}

- (void)_scanNodeTransformation:(float *)arr;
{
  NSLog(@"<%@ %x>: ASE TM information ignored...",[self class],self);
}

- (G3DASEObject *)_scanMeshData
{
  G3DASEObject *data = nil;
  int _faces = 0;
  int _vertices = 0;
  int matRef = -1;
  int loc;
    
  [self _scanIntsNamed:@"*MESH_NUMVERTEX" intoArray:&_vertices count:1];
  [self _scanIntsNamed:@"*MESH_NUMFACES" intoArray:&_faces count:1];

  if ((data = [G3DASEObject aseObjectWithFaces:_faces vertices:_vertices])) {
    loc = [_ASE_Scanner scanLocation];
    [self _scanVertexList:&data];
    [_ASE_Scanner setScanLocation:loc];
    [self _scanFaceList:&data];

    [self _scanIntsNamed:@"*MATERIAL_REF" intoArray:&matRef count:1];
    if (matRef >= 0) {
      DLOG(@"mat ref %d",matRef);
      [data setMaterialRef:matRef];
    }
    else {
      DLOG(@"No material information for the current mesh");
    }

    // Optional stuff should be scanned here
    [_ASE_Scanner setScanLocation:loc];
    [self _scanMeshNormals:&data];

    NSLog(@"<%@ %x>: ASE Object parsed: %d vertices, %d faces, material ref %d successfully scanned!",[self class],self,_vertices,_faces,matRef);
  }

  return data;
}

- (void)_scanVertexList:(G3DASEObject **)data
{
  unsigned i = 0;
  unsigned addr = 0;
  unsigned cnt = [*data vertexCount];
  float *vtx = [*data vertices];
  float p;
  int number;

  STEP_BACK(1);

  while (i<cnt) {
    if ([_ASE_Scanner isAtEnd] == NO) {      
      if ([_ASE_Scanner scanUpToString:@"*MESH_VERTEX" intoString:NULL] &&
	  [_ASE_Scanner scanString:@"*MESH_VERTEX" intoString:NULL] &&
	  [_ASE_Scanner scanInt:&number]) {

	if ([_ASE_Scanner scanFloat:&p]) vtx[addr]   = p;
	if ([_ASE_Scanner scanFloat:&p]) vtx[addr+1] = p;
	if ([_ASE_Scanner scanFloat:&p]) vtx[addr+2] = p;
	vtx[addr+3] = 1.0f;

	DLOG(@"%d: %f %f %f %f, index %d (location %d)",number,vtx[addr],vtx[addr+1],vtx[addr+2],vtx[addr+3],addr,[_ASE_Scanner scanLocation]);

	addr+=4;
	i++;

	STEP_BACK(1);
      }
    }
  }
}

- (void)_scanFaceList:(G3DASEObject **)data
{
  unsigned i;
  unsigned cnt = [*data faceCount];
  G3D_Face *faces = [*data faces];
  unsigned p;

  STEP_BACK(1);

  for (i=0;i<cnt;i++) {    
    while ([_ASE_Scanner isAtEnd] == NO) {      
      if ([_ASE_Scanner scanUpToString:@"*MESH_FACE" intoString:NULL] &&
	  [_ASE_Scanner scanString:@"*MESH_FACE" intoString:NULL]) {

	if ([_ASE_Scanner scanInt:&p]) {
	  faces[i].index = (unsigned)p;

	  if (p != i) {
	    DLOG(@"<%@ %x> Index %d should be %d",[self class],self,p,i);
	    break;
	  }
	}
	else {
	  // Happens always if i == 0 (on GNUstep) ... ?!
	  faces[i].index = i;
	}
	
	[_ASE_Scanner scanUpToString:@"A:" intoString:NULL];
	[_ASE_Scanner scanString:@"A:" intoString:NULL];
	if ([_ASE_Scanner scanInt:&p]) faces[i].a   = (unsigned)p;

	[_ASE_Scanner scanUpToString:@"B:" intoString:NULL];
	[_ASE_Scanner scanString:@"B:" intoString:NULL];
	if ([_ASE_Scanner scanInt:&p]) faces[i].b   = (unsigned)p;

	[_ASE_Scanner scanUpToString:@"C:" intoString:NULL];
	[_ASE_Scanner scanString:@"C:" intoString:NULL];
	if ([_ASE_Scanner scanInt:&p]) faces[i].c   = (unsigned)p;

	[_ASE_Scanner scanUpToString:@"AB:" intoString:NULL];
	[_ASE_Scanner scanString:@"AB:" intoString:NULL];
	if ([_ASE_Scanner scanInt:&p]) faces[i].ab   = (unsigned)p;

	[_ASE_Scanner scanUpToString:@"BC:" intoString:NULL];
	[_ASE_Scanner scanString:@"BC:" intoString:NULL];
	if ([_ASE_Scanner scanInt:&p]) faces[i].bc   = (unsigned)p;

	[_ASE_Scanner scanUpToString:@"CA:" intoString:NULL];
	[_ASE_Scanner scanString:@"CA:" intoString:NULL];
	if ([_ASE_Scanner scanInt:&p]) faces[i].ca   = (unsigned)p;

	break;
      }
    }
  }
}

- (void)_scanMeshNormals:(G3DASEObject **)data
{
  unsigned loc = [_ASE_Scanner scanLocation];

  STEP_BACK(1);

  if ([_ASE_Scanner scanUpToString:ASE_MESH_NORMALS intoString:NULL] &&
      [_ASE_Scanner scanString:ASE_MESH_NORMALS intoString:NULL]) {
    unsigned faces = [*data faceCount];
    int i = 0;
    int j;
    float *vn;

    DLOG(@"<%@ %x>: ASE Mesh Normals Found!",[self class],self);
    [*data setHasVertexNormals:YES];

    while(i<faces) {
      if([_ASE_Scanner isAtEnd] == NO) {      
	if([_ASE_Scanner scanUpToString:ASE_MESH_FNORMAL intoString:NULL] &&
	   [_ASE_Scanner scanString:ASE_MESH_FNORMAL intoString:NULL]) {

	  // The face normal is ignored for now
	  /*
	  [_ASE_Scanner scanInt:NULL];
	  [_ASE_Scanner scanFloat:NULL];
	  [_ASE_Scanner scanFloat:NULL];
	  [_ASE_Scanner scanFloat:NULL];
	  */

	  for(j=0;j<3;j++) {
	    if([_ASE_Scanner scanUpToString:ASE_MESH_VNORMAL intoString:NULL]&&
	       [_ASE_Scanner scanString:ASE_MESH_VNORMAL intoString:NULL]) {
	      unsigned vIndex;

	      // The normals have a size of 4 floats
	      if (![_ASE_Scanner scanInt:&vIndex]) {
		RAISE(@"No valid addresse scanned...");
	      }

	      vn = [*data normalAtIndex:vIndex];

	      if ([_ASE_Scanner scanFloat:vn]);
	      vn++;
	      if ([_ASE_Scanner scanFloat:vn]);
	      vn++;
	      if ([_ASE_Scanner scanFloat:vn]);

	      STEP_BACK(1);
	    }
#ifdef DEBUG
	    else {
	      NSLog(@"<%@ %x>: No Mesh Normals found!",[self class],self);
	    }
#endif DEBUG
	  }
	  i++;
	  STEP_BACK(1);
	}
	else {
	  NSLog(@"<%@ %x>: No Mesh Face Normals found!",[self class],self);
	}
      }
    } 
  }
  else {
    NSLog(@"<%@ %x>: No Mesh Face Normals found at all!",[self class],self);
    [*data setHasVertexNormals:NO];
  }
  
  [_ASE_Scanner setScanLocation:loc];
}

- (void)_scanLineNamed:(NSString *)nm intoString:(NSString **)dest
{
  STEP_BACK(1);

  while ([_ASE_Scanner isAtEnd] == NO) {
    if ([_ASE_Scanner scanUpToString:nm intoString:NULL] &&
	[_ASE_Scanner scanString:nm intoString:NULL] &&
	[_ASE_Scanner scanUpToString:@"\n" intoString:dest]) {
      return;
    }
  }
}

- (void)_scanFloatsNamed:(NSString *)nm intoArray:(float *)arr count:(int)cnt
{
  STEP_BACK(1);

  while ([_ASE_Scanner isAtEnd] == NO) {
    if ([_ASE_Scanner scanUpToString:nm intoString:NULL] &&
	[_ASE_Scanner scanString:nm intoString:NULL]) {
      int i;
      float p;

      for (i=0;i<cnt;i++) {
	if ([_ASE_Scanner scanFloat:&p]) {
	  arr[i] = p;
	}
      }
      return;
    }
  }
}

- (void)_scanIntsNamed:(NSString *)nm intoArray:(int *)arr count:(int)cnt
{
  STEP_BACK(1);

  while ([_ASE_Scanner isAtEnd] == NO) {
    if ([_ASE_Scanner scanUpToString:nm intoString:NULL] &&
	[_ASE_Scanner scanString:nm intoString:NULL]) {
      int i;
      int p;

      for (i=0;i<cnt;i++) {
	if ([_ASE_Scanner scanInt:&p]) {
	  arr[i] = p;
	}
      }
      return;
    }
  }
}

- (G3DGraphNode *)_createObjectTree
{
  NSArray *geomObjects;
  int objCount = [objects count];
  G3DASEObject *data;
  G3DMaterial *mat;
  G3DVtxArray *vertexArray;
  int _mref;

  NSAssert(objCount>0,@"No objects scanned!");

  geomObjects = [objects allValues];

  if (objCount == 1) {
    if ((data = [geomObjects objectAtIndex:0])) {
      G3DShape *node = [G3DShape shape];

      _mref = [data materialRef];
      if (_mref != -1) {
	mat = [materials objectForKey:[NSNumber numberWithInt:_mref]];
	[[node state] setMaterial:mat];
      }

      if ((vertexArray = [G3DVtxArray arrayWithASEObject:data])) {
	[node setGeometry:vertexArray];
	[node setName:[data name]];

	objects = nil;
	materials = nil;

	DLOG(@"<%@ %x>: %@ %x successfully imported!",[self class],self,[node class],node);
	return node;
      }
    }
  }
  else {
    G3DGroup *node = [G3DGroup groupWithCapacity:objCount allowReordering:YES];
    NSEnumerator *enumerator;
    id key;

    enumerator = [objects keyEnumerator];
    while ((key = [enumerator nextObject])) {
      G3DShape *leaf = [G3DShape shape];

      data = [objects objectForKey:key];

      _mref = [data materialRef];
      if (_mref != -1) {
	mat = [materials objectForKey:[NSNumber numberWithInt:_mref]];
	[[leaf state] setMaterial:mat];
      }

      if ((vertexArray = [G3DVtxArray arrayWithASEObject:data])) {
	[leaf setGeometry:vertexArray];
	[leaf setName:[data name]];
	[node addChild:leaf];
      }
    }

    objects = nil;
    materials = nil;

    DLOG(@"<%@ %x>: %@ %x successfully imported!",[self class],self,[node class],node);

    return node;
  }

  return nil;
}

@end

@implementation G3DASELoader

static G3DASELoader *_ASE_ldr = nil;

+ (id)defaultLoader
{
  if (!_ASE_ldr) {
    _ASE_ldr = [[G3DASELoader alloc] init];
  }
  return _ASE_ldr;
}

- (G3DGraphNode *)objectTreeWithContentsOfFile:(NSString *)file
{
  if (!(aseFile = [[NSString alloc] initWithContentsOfFile:file])) {
    [NSException raise:G3DASELoaderException format:@"Couldn't load %@",file];
    return nil;
  }

  if (_ASE_Scanner != nil) {
    [_ASE_Scanner release];
  }
  _ASE_Scanner = [[NSScanner alloc] initWithString:aseFile];

  return [self _parseASEFile];
}

@end






