/* 
   G3DWaterSurface.m

   Simple water animation

   Copyright (C) 2000 Philippe C.D. Robert

   Author:  Philippe C.D. Robert <robert@iam.unibe.ch>
   Date: March 2000
   
   This file is part of the interactive 3DKit

   This file is rleased under the GNU GPL!
*/ 

#import "G3DWaterSurface.h"
#import <RenderKit/RenderKit.h>

#include <time.h>

NSString *G3DWaterSurfaceException = @"G3DWaterSurfaceException";

@implementation G3DWaterSurface

float _rndVal(float ulimit)
{
  /*
   * cf Numerical Recipes in C: The Art of Scientific Computing
   *    New  York:  Cambridge  University Press,  1990 (1st ed, p. 207))
   */

  return(float)(ulimit * rand()/(RAND_MAX+1.0));
}

//-----------------------------------------------------------------------------
// Class methods
//-----------------------------------------------------------------------------

+ (G3DWaterSurface *)surfaceWithSlices:(int)sl scale:(float)sc
{
  return [[[G3DWaterSurface alloc] initWithSlices:sl scale:sc] autorelease];
}

//-----------------------------------------------------------------------------
// Init and free
//-----------------------------------------------------------------------------

- (id)initWithSlices:(int)sl scale:(float)sc
{
  if ((self = [super init])) {
    unsigned  n;

    srand((unsigned)time(NULL));

    ang_x = 50.0f;
    _sl = sl;
    _f = 1.0f/_sl;

    scale = sc;
 
    inc_ang_x = 0.0f;
    inc_ang_y = 0.0f;

    hasWaves = NO;

    n = _sl * _sl * sizeof(float);
    water1 = malloc(n);
    water2 = malloc(n);
    water3 = malloc(n);

    if (water1 != NULL && water2 != NULL && water3 != NULL) {
      memset(water1,0,n);
      memset(water2,0,n);
      memset(water3,0,n);
    }
    else {
      free(water1);
      free(water2);
      free(water3);

      [NSException raise:G3DWaterSurfaceException format:@"Could not initialise the water array!"];
    }

    stage[0] = water1;
    stage[1] = water2;
    stage[2] = water3;

    altitude = 0.02;
    focusVelocity = 0.06;
    lowering = 0.995f;
    dropRate = 200;
  }
  return self;
}

- (void)dealloc
{
  if ([self retainCount] == 1) {
    free(water1);
    free(water2);
    free(water3);
  }

  [super dealloc];
}

//-----------------------------------------------------------------------------
// Water rendering
//-----------------------------------------------------------------------------

- (void)updateWater
{
  int i,j;

  int first;
  int second;
  int last;

  float *w1;
  float *w2;
  float *w3;

  ang_x+=inc_ang_x;
  ang_y+=inc_ang_y;
  
  if (hasWaves) {
    int x = (int)_sl/2+(int)(_sl/3*cos(ang_circle));
    int y = (int)_sl/2+(int) (_sl/3*sin(ang_circle));
    int w = 5;
    int h = 5;
    float height = altitude * 0.5f;

    /*
     * elevate a box at [x,y,x+w,y+h] over the stage "layer" of the grid
     */

    w1 = stage[current];
    
    for (j=y; j<=y+h; j++) {
      for (i=x; i<=x+w; i++) {
	//water[i%(_sl)][j%(_sl)][current] += height;
	// vertex at node [i%_sl][j%_sl]
	*(w1 + (_sl * (i % _sl) + (j % _sl))) += height;
      }
    }
    ang_circle += focusVelocity;
  }
  
  // calculate_water(self);
  /*
   * Integration step
   *
   */

  first=current;
  second=(current+1)%3;
  last=(current+2)%3;

  w1 = stage[first];
  w2 = stage[second];
  w3 = stage[last];
  
  for (i=1; i<_sl-1; i++) {
    for (j=1; j<_sl-1; j++) {
      //w3[i][j] = -w1[i][j]+0.5*(w2[i-1][j]+w2[i+1][j]+w2[i][j-1]+w2[i][j+1]);
      *(w3+(_sl*i)+j) = -(*(w1+(_sl*i)+j)) + 0.5 * 
	(*(w2+(_sl*(i-1))+j) + 
	 *(w2+(_sl*(i+1))+j) + 
	 *(w2+(_sl*i)+(j-1)) +
	 *(w2+(_sl*i)+(j+1)));
    }
  }
  
  // calm_water(self,(current+1)%3);
  
  w1 = stage[(current+1)%3];
  w2 = stage[(current+2)%3];
  
  for (i=0; i<_sl; i++) {
    int _fi = i * _sl;

    for (j=0; j<_sl; j++) {
      int _fij = _fi + j;

      *(w1+_fij) *= lowering;
      *(w2+_fij) *= lowering;
    }
  }
  
   current=(current+1)%3;

   [self animate];
}

- (void)animate
{
  static int wc = 0;

  if (wc == 0) {
    int i,j;
    
    int x = (int)_rndVal(_sl);
    int y = (int)_rndVal(_sl);
    int w = (int)_rndVal(10);
    int h = (int)_rndVal(10);
    float height = _rndVal(altitude*5);

    float *wp = stage[current];

    /*
     * elevate a box at [x,y,x+w,y+h] over the stage "layer" of the grid
     */
    
    for (j=y; j<=y+h; j++) {
      int _ji = j%_sl;

      for (i=x; i<=x+w; i++) {
	*(wp+(_sl*(i%_sl))+_ji) += height;
      }
    }
  }
  
  wc++;
  if (wc > dropRate) {
    wc = 0;
  }   
}

- (void)setHasWaves:(BOOL)yn
{
  hasWaves = yn;
}

- (BOOL)hasWaves
{
  return hasWaves;
}

- (void)setDropRate:(int)val
{
  dropRate = val;
}

- (int)dropRate
{
  return dropRate;
}

- (void)setLowering:(float)val
{
  lowering = val;
}

- (float)lowering
{
  return lowering;
}

- (void)setAltitude:(float)val
{
  altitude = val;
}

- (float)altitude
{
  return altitude;
}

//-----------------------------------------------------------------------------
// Geometry backend
//-----------------------------------------------------------------------------

- (void)drawGeometry
{
  register int i,j;
  float *w = stage[current];
  
  G3D_PUSH_MATRIX;

  glFrontFace(GL_CW);
  
  glScalef(scale,1.0 + (1.0/scale),scale);
  glTranslatef(-0.5f,0.0f,0.5f);
  
  for (j=1; j<_sl-1; j++) {
    glBegin(GL_TRIANGLE_STRIP);
    
    for (i=1; i<_sl-1; i++) {
      // float col1 = water[i][j][current]+1;
      float col1 = *(w+(i*_sl)+j)+1;
      // float col2 = water[i][j+1][current]+1;
      float col2 = *(w+(i*_sl)+j+1)+1;
      float _imf = (float)i * _f;

      glColor3f(col1,col1,col1);
      glTexCoord2f((float)i/_sl+((*(w+(i*_sl)+j)-*(w+((i+1)*_sl)+j))/_f)/35,
		   (float)j/_sl+((*(w+(i*_sl)+j) - *(w+(i*_sl)+j+1))/_f)/35);
      glVertex3f(_imf,*(w+((i*_sl)+j)),(float)-j*_f);
      
      glColor3f(col2,col2,col2);
      glTexCoord2f((float)i/_sl+((*(w+(i*_sl)+j+1) - *(w+((i+1)*_sl)+j+1))/_f)/35.0f,
		   (float)(j+1)/_sl+((*(w+(i*_sl)+j+1) - *(w+(i*_sl)+j+2))/_f)/35);
      glVertex3f(_imf,*(w+((i*_sl)+j+1)),(float)-(j+1)*_f);
    }
    glEnd();
  }

  glFrontFace(GL_CCW);

  G3D_POP_MATRIX;

  [self updateWater];
}

- (void)precompile
{
  DLOG(@"%@ %x cannot be precompiled!",[self class],self);
}

- (int)vertexCount
{
  return -1;
}

- (void)getMin:(float *)min max:(float *)max
{
  /*
   * To be refined - this is a QuickHack!!!
   *
   */

  min[0] = (float)-0.5f*scale;
  min[1] = (float)-G3DEPSILON;
  min[2] = (float)-0.5*scale;

  max[0] = (float)0.5f*scale;
  max[1] = (float)G3DEPSILON;
  max[2] = (float)0.5f*scale;
}

- (void)setShape:(G3DShape *)aShape
{
  shape = aShape;
}

//-----------------------------------------------------------------------------
// NSCoding
//-----------------------------------------------------------------------------

- (void)encodeWithCoder:(NSCoder *)aCoder
{
  [aCoder encodeObject:shape];
  [aCoder encodeValueOfObjCType:@encode(int) at:&_sl];
}

- (id)initWithCoder:(NSCoder *)aCoder
{
  self = [super init];
  shape = [[aCoder decodeObject] retain];
  [aCoder decodeValueOfObjCType:@encode(int) at:&_sl];
  return self;
}

//-----------------------------------------------------------------------------
// NSCopying
//-----------------------------------------------------------------------------

- (id)copyWithZone:(NSZone *)zone
{
  G3DWaterSurface *copy = [[G3DWaterSurface allocWithZone:zone] init];

  copy->shape = shape;
  copy->_sl = _sl;

  return copy;
}

@end









