/* 
   GlutDemo.m

   Simple rendering demo based on the glut camera class

   Copyright (C) 2000 Philippe C.D. Robert

   Author:  Philippe C.D. Robert <robert@iam.unibe.ch>
   Date: March 2000
   
   This file is part of the GNU 3DKit

   This file is rleased under the GNU GPL!
*/ 

#import <Foundation/Foundation.h>
#import <RenderKit/RenderKit.h>
#import <GlutKit/G3DGlutKit.h>
#import "G3DShape+WaterSurface.h"

#define STEP 0.3
#define ANGLE 3.0f

#define WATER_TEX @"water.bmp"

@interface AppController : NSObject
{
  G3DGlutCamera *cam;
  G3DWorld *world;
  G3DShape *water;
}

- (void)applicationDidFinishLaunching;
- (void)dealloc;

- (void)cameraDidRenderFrame:(id)sender;

- (void)glutCamera:(id)sender 
   didReceiveEvent:(unsigned char)c 
		 x:(int)x 
		 y:(int)y;

@end

@implementation AppController

- (void)applicationDidFinishLaunching
{
  NSAutoreleasePool *pool;
  G3DLight *lightA;
  G3DColour *c;
  G3DGroup *tg;
  G3DTexture *tex;
  G3DWrapping wrp;
  G3DVector3f *vec;
  G3DVector3f *lpos;
  G3DVector3f *dir;

  float pos[3] = {0,4,10};
  float to[3] = {0,0,0};
  float up[3] = {0,1,0};

  // Setting up the camera and the scenegraph
  world = [[G3DWorld alloc] init];

  NSLog(@"Initialising the camera and GL...");

  cam = [[G3DGlutCamera alloc] initWithSize:NSMakeSize(640,480) fullscreen:NO];
  DLOG(@"Camera %x of class %@ successfully instantiated!",cam,[cam class]);

  [cam setRenderDelegate:self];
  [cam setWorld:world];
  [cam setUseTwoSideModel:NO];
  [cam setEnableLighting:YES];
  [cam setupCamera];

  // =====================================
  // The lighting
  // =====================================

  NSLog(@"Initialising 1 global light...");

  pool = [[NSAutoreleasePool alloc] init];

  lightA = [G3DLight pointLightWithPosition:[G3DVector3f vector3fWithX:1.5f 
							 y:2.0f 
							 z:0.0f]];
  [lightA setDiffuse:[G3DColour colourWithR:0.8 g:0.0 b:0.7 a:1.0]];
  [lightA setAmbient:[G3DColour colourWithR:0.2 g:0.9 b:0.7 a:1.0]];
  [lightA setSpecular:[G3DColour colourWithR:0.75 g:1.0 b:0.75 a:1.0]];
  [world addDynamicObject:lightA];

  lpos = [G3DVector3f vector3fWithX:2.0f y:2.0f z:2.0f];
  dir = [G3DVector3f vector3fWithX:-1.0f y:-1.0f z:-1.0f];

  lightA = [G3DLight spotLightWithPosition:lpos direction:dir];
  
  [lightA setDiffuse:[G3DColour colourWithR:0.4 g:0.2 b:0.7 a:1.0]];
  [lightA setAmbient:[G3DColour colourWithR:0.7 g:0.4 b:0.3 a:1.0]];
  [lightA setSpecular:[G3DColour colourWithR:1.0 g:1.0 b:1.0 a:1.0]];
  [world addDynamicObject:lightA];

  // =====================================
  // The scenegraph structure
  // =====================================

  NSLog(@"Setting up the scenegraph...");

  tg = [G3DGroup group];

  NS_DURING
  tex = [G3DTexture textureWithContentsOfFile:WATER_TEX];
  [tex setBuildMipmaps:YES];
  NS_HANDLER
  [tex autorelease];
  tex = nil;
  NSLog([localException description]);
  NS_ENDHANDLER

  water = [G3DShape waterSurfaceWithSlices:10 scale:10 texture:tex];
  [water setEnableLighting:YES];
  [water setEnableBlending:YES];
  [[water geometry] setHasWaves:NO];
  [tg addChild:water];

  [world addDynamicObject:tg];

  // =====================================
  // Fog
  // =====================================

  [cam setFog:[G3DFog fog]];

  [cam setCamera:pos toward:to up:up];
  [cam preprocessScene];

  [pool release];

  [cam renderScene];
}

- (void)cameraDidRenderFrame:(id)sender
{
  // Add you stuff here

  glutPostRedisplay();
}

- (void)glutCamera:(id)sender didReceiveEvent:(unsigned char)c x:(int)x y:(int)y
{
  static BOOL l = 0;

  switch (c) {
  case 27:
    exit(0);
    break;
  case ' ':
    break;
  case '2':
    [sender moveX:0.0f y:0.0f z:STEP];
    break;
  case '4':
    [sender moveX:-STEP y:0.0f z:0.0f];
    break;
  case '6':
    [sender moveX:STEP y:0.0f z:0.0f];
    break;
  case '8':
    [sender moveX:0.0f y:0.0f z:-STEP];
    break;
  case 'q':
    [sender moveX:0.0f y:STEP z:0.0f];
    break;
  case 'w':
    [sender moveX:0.0f y:-STEP z:0.0f];
    break;
  case 'a':
    [sender rotateY:-ANGLE];
    break;
  case 's':
    [sender rotateY:ANGLE];
    break;
  case 'd':
    [sender rotateX:-ANGLE];
    break;
  case 'f':
    [sender rotateX:ANGLE];
    break;
  case 'g':
    [sender rotateZ:-ANGLE];
    break;
  case 'h':
    [sender rotateZ:ANGLE];
    break;
  case 'l':
    [sender setEnableLighting:l];
    l = !l;
    break;
  default:
    return;
  }

  glutPostRedisplay();
}

- (void)dealloc
{
  /*
   * This won't ever be invoked, I guess....;-)
   */

  [[G3DRenderEngine renderEngine] cleanup];

  [cam release];
  [world release];
  [water release];

  [super dealloc];
}

@end

int main(int argc, char **argv)
{
  AppController *controller;
  id pool = [NSAutoreleasePool new];

  // Since we use the glut based camera, we should initialise glut first!
  glutInit(&argc, argv);

  // The delegate
  controller = [[AppController alloc] init];
  [controller applicationDidFinishLaunching];

  // The end...
  [controller release];
  [pool release];

  return 0;
}




